/* List of defines generated by configure. Included with preprocessor flag,
 * -include, to avoid long list of -D defines on the compile command-line.
 * Do not edit.
 */

#ifndef js_confdefs_h
#define js_confdefs_h

#define CPP_THROW_NEW throw()
#define D_INO d_ino
#define EDITLINE 1
#define HAVE_64BIT_BUILD 1
#define HAVE_ACOSH 1
#define HAVE_ASINH 1
#define HAVE_ATANH 1
#define HAVE_CBRT 1
#define HAVE_CPP_AMBIGUITY_RESOLVING_USING 1
#define HAVE_CPP_DYNAMIC_CAST_TO_VOID_PTR 1
#define HAVE_DIRENT_H 1
#define HAVE_DLOPEN 1
#define HAVE_ENDIAN_H 1
#define HAVE_EXPM1 1
#define HAVE_GETC_UNLOCKED 1
#define HAVE_GETOPT_H 1
#define HAVE_GMTIME_R 1
#define HAVE_GNU_LIBC_VERSION_H 1
#define HAVE_I18N_LC_MESSAGES 1
#define HAVE_LANGINFO_CODESET 1
#define HAVE_LIBM 1
#define HAVE_LINUX_QUOTA_H 1
#define HAVE_LOCALECONV 1
#define HAVE_LOCALECONV 1
#define HAVE_LOCALTIME_R 1
#define HAVE_LOG1P 1
#define HAVE_LOG2 1
#define HAVE_MALLOC_H 1
#define HAVE_MALLOC_USABLE_SIZE 1
#define HAVE_MBRTOWC 1
#define HAVE_MEMORY_H 1
#define HAVE_NL_TYPES_H 1
#define HAVE_POSIX_FADVISE 1
#define HAVE_POSIX_FALLOCATE 1
#define HAVE_RES_NINIT 1
#define HAVE_SETLOCALE 1
#define HAVE_SSIZE_T 1
#define HAVE_SYS_BITYPES_H 1
#define HAVE_SYS_CDEFS_H 1
#define HAVE_SYS_MOUNT_H 1
#define HAVE_SYS_QUOTA_H 1
#define HAVE_SYS_STATFS_H 1
#define HAVE_SYS_STATVFS_H 1
#define HAVE_SYS_VFS_H 1
#define HAVE_THREAD_TLS_KEYWORD 1
#define HAVE_TM_ZONE_TM_GMTOFF 1
#define HAVE_TRUNC 1
#define HAVE_UNISTD_H 1
#define HAVE_VA_COPY 1
#define HAVE_VISIBILITY_ATTRIBUTE 1
#define HAVE_VISIBILITY_HIDDEN_ATTRIBUTE 1
#define HAVE_WCRTOMB 1
#define HAVE___CXA_DEMANGLE 1
#define JS_CODEGEN_NONE 1
#define JS_DEFAULT_JITREPORT_GRANULARITY 3
#define JS_HAVE_ENDIAN_H 1
#define JS_POSIX_NSPR 1
#define JS_PUNBOX64 1
#define JS_STANDALONE 1
#define MALLOC_H <malloc.h>
#define MOZILLA_UAVERSION "38.0"
#define MOZILLA_VERSION "38.6.1"
#define MOZILLA_VERSION_U 38.6.1
#define MOZJS_MAJOR_VERSION 38
#define MOZJS_MINOR_VERSION 6
#define MOZ_DEBUG_SYMBOLS 1
#define MOZ_DLL_SUFFIX ".so"
#define MOZ_UPDATE_CHANNEL default
#define NS_ATTR_MALLOC __attribute__((malloc))
#define NS_WARN_UNUSED_RESULT __attribute__((warn_unused_result))
#define RELEASE_BUILD 1
#define STDC_HEADERS 1
#define VA_COPY va_copy
#define XP_UNIX 1
#define X_DISPLAY_MISSING 1
#define _REENTRANT 1

#include "js/RequiredDefines.h"

#endif /* js_confdefs_h */
