/*-
 * Copyright (c) 2014-2016 MongoDB, Inc.
 * Copyright (c) 2008-2014 WiredTiger, Inc.
 *	All rights reserved.
 *
 * See the file LICENSE for redistribution information.
 */

#define	DUMP_JSON_VERSION_MARKER	"WiredTiger Dump Version"
#define	DUMP_JSON_CURRENT_VERSION	1
#define	DUMP_JSON_SUPPORTED_VERSION	1
