/* WiredTiger-specific styling. */ 
body, table, div, p, dl {
	font-family: "Helvetica Neue", Helvetica, Arial, sans-serif;
	color: #373737;
	font-size: 13px;
	line-height: 1.3;
}

h1 {
	font-size: 120%;
}

h2 {
	font-size: 110%;
}

/* @group Link Styling */

a {
	color: #3D578C;
}

.tablist {
	width: 800px;
}

.tablist a[href="pages.html"] {
	display: none;
}

.contents a:visited {
	color: #4665A2;
}

#projectlogo {
	vertical-align: middle;
}
 
.banner {
	position: absolute; top: -2px; right: 1em;
	overflow: hidden;
	background-color: #e57504;
	background-image: -webkit-linear-gradient(top, #dc7202, #ee7906);
	background-image: -moz-linear-gradient(top, #dc7202, #ee7906);
	background-image: -ms-linear-gradient(top, #dc7202, #ee7906);
	border-radius: 0 0 8px 8px;
	-moz-border-radius: 0 0 8px 8px;
	box-shadow: 0 0 1em #888;
}

.banner a {
	border: 1px solid #ebaa65;
	border-top-style: none;
	color: #fff;
	display: block;
	font: bold 9pt sans-serif;
	margin: 0.05em 0 0.05em 0;
	padding: 0.5em 2em;
	text-align: center;
	text-decoration: none;
	text-shadow: 0 0 0.5em #444;
}

.banner a.last {
	border: none;
}

@media print
{
  .banner { display: none; }
}
