/**
 *    Copyright (C) 2015 MongoDB Inc.
 *
 *    This program is free software: you can redistribute it and/or  modify
 *    it under the terms of the GNU Affero General Public License, version 3,
 *    as published by the Free Software Foundation.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Affero General Public License for more details.
 *
 *    You should have received a copy of the GNU Affero General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    As a special exception, the copyright holders give permission to link the
 *    code of portions of this program with the OpenSSL library under certain
 *    conditions as described in each individual source file and distribute
 *    linked combinations including the program with the OpenSSL library. You
 *    must comply with the GNU Affero General Public License in all respects for
 *    all of the code used other than as permitted herein. If you modify file(s)
 *    with this exception, you may extend this exception to your version of the
 *    file(s), but you are not obligated to do so. If you do not wish to do so,
 *    delete this exception statement from your version. If you delete this
 *    exception statement from all source files in the program, then also delete
 *    it in the license file.
 */

#pragma once

#include "mongo/db/matcher/extensions_callback.h"

namespace mongo {

class NamespaceString;
class OperationContext;

/**
 * ExtensionsCallbackReal uses the provided OperationContext and namespace to capture context
 * necessary for parsing $text and $where clauses.
 */
class ExtensionsCallbackReal : public ExtensionsCallback {
public:
    /**
     * Does not take ownership of 'nss' or 'txn'.
     *
     * 'nss' must outlive this object. 'txn' must outlive this object also; in addition, 'txn' must
     * outlive any MatchExpression objects generated by these callbacks (as the generated objects
     * are allowed to keep a reference to 'txn').
     */
    ExtensionsCallbackReal(OperationContext* txn, const NamespaceString* nss);

    /**
     * Returns a TextMatchExpression, or an error Status if parsing fails.
     */
    StatusWithMatchExpression parseText(BSONElement text) const final;

    /**
     * Returns a WhereMatchExpression, or an error Status if parsing fails.
     */
    StatusWithMatchExpression parseWhere(BSONElement where) const final;

private:
    OperationContext* const _txn;
    const NamespaceString* const _nss;
};

}  // namespace mongo
